
#include "platform.h"
#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>

#include <linux/input.h>
#include <X11/keysym.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


extern "C" {
#include "linux/omapfb.h" //
#include "linux/fbdev.h"
#include "linux/xenv.h"
}


extern CFG_OPTIONS config_options;

unsigned short *VideoBuffer = NULL;
static const int G_Buffers = 4;

static struct vout_fbdev *fb;

static int omap_setup_layer(int fd, int enabled, int x, int y, int w, int h, int first_call)
{
  struct omapfb_plane_info pi = { 0, };
  struct omapfb_mem_info mi = { 0, };
  int ret;

  ret = ioctl(fd, OMAPFB_QUERY_PLANE, &pi);
  if (ret != 0) {
    perror("QUERY_PLANE");
    return -1;
  }

  ret = ioctl(fd, OMAPFB_QUERY_MEM, &mi);
  if (ret != 0) {
    perror("QUERY_MEM");
    return -1;
  }

  /* must disable when changing stuff */
  if (pi.enabled) {
    pi.enabled = 0;
    ret = ioctl(fd, OMAPFB_SETUP_PLANE, &pi);
    if (ret != 0)
      perror("SETUP_PLANE");
  }

  if (first_call) {
    mi.size = 640*512*3*3;
    ret = ioctl(fd, OMAPFB_SETUP_MEM, &mi);
    if (ret != 0) {
      perror("SETUP_MEM");
      return -1;
    }
  }

  pi.pos_x = x;
  pi.pos_y = y;
  pi.out_width = w;
  pi.out_height = h;
  pi.enabled = enabled;

  ret = ioctl(fd, OMAPFB_SETUP_PLANE, &pi);
  if (ret != 0) {
    perror("SETUP_PLANE");
    return -1;
  }

  return 0;
}


static void set_filter(int is_filtered)
{
  static int was_filtered = -1;
  char buf[128];

  if (is_filtered == was_filtered)
    return;

  snprintf(buf, sizeof(buf), "sudo -n /usr/pandora/scripts/op_videofir.sh %s",
    is_filtered ? "default" : "none");
  system(buf);
  was_filtered = is_filtered;
}

void set_video_mode(int w, int h, int buffers, int scale, int filter)
{
  int lx, ly, lw = w, lh = h;
  switch (scale) {
    case 0:
      lw = w;
      lh = h;
      break;
    case 1:
      lw = w * 2;
      lh = h * 2;
      break;
    case 2:
      lw = 800;
      lh = 480;
      break;
    default:
      printf( "unknown scale: %d\n", scale);
      break;
  }
  if (lw > 800)
    lw = 800;
  if (lh > 480)
    lh = 480;

  lx = 800 / 2 - lw / 2;
  ly = 480 / 2 - lh / 2;

  omap_setup_layer(vout_fbdev_get_fd(fb), 1, lx, ly, lw, lh, 0);
  set_filter(filter);

  vout_fbdev_resize(fb, w, h, 16, 0, 0, 0, 0, buffers);
}

/**/

#define PND_UP         XK_Up
#define PND_DOWN       XK_Down
#define PND_LEFT       XK_Left
#define PND_RIGHT      XK_Right
#define PND_SELECT     XK_Alt_L
#define PND_START      XK_Control_L
#define PND_TRIGGER_L  XK_Shift_R
#define PND_TRIGGER_R  XK_Control_R
#define PND_KEYPAD_A   XK_Home
#define PND_KEYPAD_B   XK_End
#define PND_KEYPAD_X   XK_Page_Down
#define PND_KEYPAD_Y   XK_Page_Up


struct keymap
{
    unsigned int xk;
    unsigned int fba;
};

static keymap s_keymap_normal[] = {
    { PND_UP        ,  FBA_KEYPAD_UP     },
    { PND_DOWN      ,  FBA_KEYPAD_DOWN   },
    { PND_LEFT      ,  FBA_KEYPAD_LEFT   },
    { PND_RIGHT     ,  FBA_KEYPAD_RIGHT  },
    { PND_SELECT    ,  FBA_KEYPAD_COIN   }, //select
    { PND_START     ,  FBA_KEYPAD_START  }, //start
    { PND_KEYPAD_A  ,  FBA_KEYPAD_A      }, //a
    { PND_KEYPAD_B  ,  FBA_KEYPAD_C      }, //b
    { PND_KEYPAD_X  ,  FBA_KEYPAD_B      }, //x
    { PND_KEYPAD_Y  ,  FBA_KEYPAD_D      }, //y
    { PND_TRIGGER_L ,  FBA_KEYPAD_E      }, //is trigger L
    { PND_TRIGGER_R ,  FBA_KEYPAD_F      }, //is trigger R
    {0,0}
};

static keymap s_keymap_vertical[] = {
    { PND_UP        ,  FBA_KEYPAD_LEFT   },
    { PND_DOWN      ,  FBA_KEYPAD_RIGHT  },
    { PND_LEFT      ,  FBA_KEYPAD_DOWN   },
    { PND_RIGHT     ,  FBA_KEYPAD_UP     },
    { PND_SELECT    ,  FBA_KEYPAD_COIN   }, //select
    { PND_START     ,  FBA_KEYPAD_START  }, //start
    { PND_KEYPAD_A  ,  FBA_KEYPAD_A      }, //a
    { PND_KEYPAD_B  ,  FBA_KEYPAD_C      }, //b
    { PND_KEYPAD_X  ,  FBA_KEYPAD_B      }, //x
    { PND_KEYPAD_Y  ,  FBA_KEYPAD_D      }, //y
    { PND_TRIGGER_L ,  FBA_KEYPAD_E      }, //is trigger L
    { PND_TRIGGER_R ,  FBA_KEYPAD_F      }, //is trigger R
    {0,0}
};

#define STR_TO_KEY( k ) if ( strstr(key,#k)!=0 ) { *keyout = k; return true; }
static bool StringToKeyFBA(const char* key, unsigned int* keyout )
{
    STR_TO_KEY(FBA_KEYPAD_UP)
    STR_TO_KEY(FBA_KEYPAD_DOWN)
    STR_TO_KEY(FBA_KEYPAD_LEFT)
    STR_TO_KEY(FBA_KEYPAD_RIGHT)
    STR_TO_KEY(FBA_KEYPAD_START)
    STR_TO_KEY(FBA_KEYPAD_COIN)
    STR_TO_KEY(FBA_KEYPAD_A)
    STR_TO_KEY(FBA_KEYPAD_B)
    STR_TO_KEY(FBA_KEYPAD_C)
    STR_TO_KEY(FBA_KEYPAD_D)
    STR_TO_KEY(FBA_KEYPAD_E)
    STR_TO_KEY(FBA_KEYPAD_F)
    printf("Cannot convert Key '%s'\n",key);
    return false;
}

static bool StringToKeyPND(const char* key, unsigned int* keyout )
{
    STR_TO_KEY(PND_UP)
    STR_TO_KEY(PND_DOWN)
    STR_TO_KEY(PND_LEFT)
    STR_TO_KEY(PND_RIGHT)
    STR_TO_KEY(PND_START)
    STR_TO_KEY(PND_SELECT)
    STR_TO_KEY(PND_KEYPAD_A)
    STR_TO_KEY(PND_KEYPAD_B)
    STR_TO_KEY(PND_KEYPAD_X)
    STR_TO_KEY(PND_KEYPAD_Y)
    STR_TO_KEY(PND_TRIGGER_L)
    STR_TO_KEY(PND_TRIGGER_R)
    printf("Cannot convert Key '%s'\n",key);
    return false;
}
#undef STR_TO_KEY

static void changeKey( unsigned int fbkey, unsigned int xk )
{
    int i;

    i = 0;
    while (s_keymap_normal[i].xk!=0)
    {
        if ( s_keymap_normal[i].fba==fbkey )
        {
            s_keymap_normal[i].xk = xk;
            break;
        }
        i++;
    }

    i = 0;
    while (s_keymap_vertical[i].xk!=0)
    {
        if ( s_keymap_vertical[i].fba==fbkey )
        {
            s_keymap_vertical[i].xk = xk;
            break;
        }
        i++;
    }
}

#define isspace(k) (k==' ' || k=='\t' || k=='\r' || k=='\n')

static char* trim( char* str )
{
    if (str==0 || str[0]==0)
        return str;

    while(isspace(*str)) str++;

    if (*str == 0)
        return str;

    char *end = str + strlen(str) - 1;
    while(end > str && isspace(*end))
        end--;
   *(end+1) = 0;
    return str;
}

static void changeKeyMap( )
{
    if ( config_options.option_keymap[0]==0 )
        return;

    printf("changeKeyMap: '%s'\n",config_options.option_keymap);

    FILE* f = fopen(config_options.option_keymap,"rt");
    if ( !f )
    {
        printf("changeKeyMap: '%s' not found.\n",config_options.option_keymap);
        return;
    }

    char line[200+1];

    while ( fgets(line,200,f)!=NULL )
    {
        if ( line[0]=='#' )
            continue;
        if ( strchr(line,'=')==0 )
            continue;

//      printf("keymap line '%s'\n",line);

        char* fbastr = strtok(line,"=");
        char* pndstr = strtok(NULL,"=");
        fbastr = trim(fbastr);
        pndstr = trim(pndstr);

        if ( fbastr!=0 && pndstr!=0 )
        {
            unsigned int fba=0;
            if ( StringToKeyFBA(fbastr,&fba)==false )
                continue;
            unsigned int pnd=0;
            if ( StringToKeyPND(pndstr,&pnd)==false )
                continue;

            printf("Changed KeyMapping: %s(%d) -> %s(%d)\n",pndstr,pnd,fbastr,fba);

            changeKey(fba,pnd);
        }
    }
    fclose(f);
}


/**/

void platform_initialize()
{
	printf("enter - platform_initialize\n");

    changeKeyMap();

    int ret, w, h, fd;
    const char *layer_fb_name;

    layer_fb_name = getenv("FBDEV_LAYER");
    if (layer_fb_name == NULL)
        layer_fb_name = "/dev/fb1";

    ret = xenv_init();
    if (ret != 0) {
        printf( "xenv_init failed with %d\n", ret);
        exit(1);
    }

    // must set the layer up first to be able to use it
    fd = open(layer_fb_name, O_RDWR);
    if (fd == -1) {
        printf( "%s: ", layer_fb_name);
        perror("open");
        exit(1);
    }

    ret = omap_setup_layer(fd, 0, 0, 0, 384, 224, 1);
    close(fd);
    if (ret != 0) {
        printf( "failed to set up layer, exiting.\n");
        exit(1);
    }

    w = 320;
    h = 240;
    fb = vout_fbdev_init("/dev/fb1", &w, &h, 16, 4);
    if (fb == NULL) {
        printf( "vout_fbdev_init failed\n");
        exit(1);
    }

    set_video_mode(w,h,G_Buffers,1,0);

    VideoBuffer = (unsigned short*)vout_fbdev_flip(fb);

	printf("leave - platform_initialize\n");
}

void platform_terminate(char *frontend)
{
    xenv_finish();
    omap_setup_layer(vout_fbdev_get_fd(fb), 0, 0, 0, 0, 0, 0);
    vout_fbdev_finish(fb);

	fcloseall();
  	sync();
}

void platform_setvideo_mode(int w, int h)
{
	printf("w: %d h: %d\n",w,h);

	set_video_mode(w,h,G_Buffers,config_options.option_scale,config_options.option_filter);

	VideoBuffer = (unsigned short*)vout_fbdev_flip(fb);
}

// input handling

static char S_platform_request_service = 0;
static char S_platform_p1p2_start = 0;
static char S_platform_exit = 0;
static char S_platform_pause = 0;


int platform_numinputs()
{
    return 1;
}

void platform_input_update(int inputId, int vertical, unsigned int* fbaKeyState)
{
    if (inputId!=0)
    {

        *fbaKeyState = 0;
        return;
    }

    static unsigned long gkeystate;
    int key, is_down, i;
    int gkey = -1;

    key = xenv_update(&is_down);

//    S_platform_p1p2_start = 0;
//    S_platform_request_service = 0;

    if (key=='p')
    {
        if ( is_down )
            S_platform_pause ^= 1;
    }
    else
    if (key=='f')
    {
        if ( is_down )
        {
            extern void ChangeFrameskip();
            ChangeFrameskip();
        }
    }
    if (key==XK_space)
    {
        if ( is_down )
            S_platform_request_service = 1;
        else
            S_platform_request_service = 0;
    }
    else
    if (key == XK_Escape)
    {
        if ( is_down )
            S_platform_exit = 1;
    }
    else
    if (key== XK_Return)
    {
        if ( is_down )
            S_platform_p1p2_start = 1;
        else
            S_platform_p1p2_start = 0;
    }

    const keymap* map = vertical ? s_keymap_vertical : s_keymap_normal;

    while ( map->xk!=0 )
    {
        if (key == map->xk)
        {
            if (is_down)
                *fbaKeyState |=   map->fba;
            else
                *fbaKeyState &= ~(map->fba);

            break;
        }
        map ++;
    }

}

char platform_request_service()
{
    return S_platform_request_service;
}

char platform_p1p2_start()
{
    return S_platform_p1p2_start;
}

char platform_exit()
{
    return S_platform_exit;
}

char platform_pause()
{
    return S_platform_pause;
}

void platform_video_flip()
{
    if ( config_options.option_vsync )
    {
        vout_fbdev_wait_vsync(fb);
    }

    VideoBuffer = (unsigned short*)vout_fbdev_flip(fb);
}

void platform_clear_framebuffers()
{
    vout_fbdev_clear(fb);
}
